/**********************************************************************
* 
* regionlock.h
*
* This file is part of MitoGen
* 
* Copyright (C) 2013-2017 -- Centre for Biomedical Image Analysis (CBIA)
* http://cbia.fi.muni.cz/
* 
* MitoGen is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* MitoGen is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with MitoGen. If not, see <http://www.gnu.org/licenses/>.
* 
* Author: David Wiesner
* 
* Description: Definition of methods of class 'SceneLocks' that handles
* region-based locks in the multi-threaded implementation of MitoGen.
*
***********************************************************************/
#ifdef GTGEN_WITH_MULTITHREADING

#ifdef GTGEN_WITH_LOCKVISUALIZATION
	#include <i3d/transform.h> // i3d::Resample
#endif

#include "regionlock.h"


std::string GetSceneTypeStr(const scene_type &sceneType)
{
	std::string scTypeStr;
	
	switch (sceneType)
	{
		case all:
			scTypeStr = "all";
			break;
		case masks:
			scTypeStr = "masks";
			break;
		case nuclei:
			scTypeStr = "nuclei";
			break;
		case phantoms:
			scTypeStr = "phantoms";
				break;
		case territories:
			scTypeStr = "territories";
			break;
		default:
			throw ERROR_REPORT("Undefined scene type!");
	}

	return scTypeStr;
}


lock_handle SceneLocks::checkAndLock(CustomLock* newLock)
{
#ifdef MITOGEN_DEBUG_MULTITHREADING
	// get start time
	const report_clock::time_point startTime = report_clock::now();	
#endif	
				
	bool lockAcquired = false;

	std::unique_lock<std::mutex> listSearchLock(listSearchLock_st);

	std::list<CustomLock*>::iterator iter;

	while (!lockAcquired)
	{
		iter = listOfCustomLocks.begin();
	
		bool alreadyLocked = false;
		
		// go over all active locks
		while (iter != listOfCustomLocks.end())
		{	
			if (((*iter)->getSceneType() == newLock->getSceneType())
				|| ((*iter)->getSceneType() == all)
				|| (newLock->getSceneType() == all))
			{
				switch (newLock->getLockType())
				{
					// newLock type is timePoint -------------------------------
					case timePoint_lock:
						if (!ableToLockTimePoints(*iter, newLock))
							alreadyLocked = true;			
						break;
					
					// newLock type is sphere ----------------------------------
					case sphere_lock:
						switch ((*iter)->getLockType())
						{
							case timePoint_lock:
								if (!ableToLockTimePoints(*iter, newLock))
									alreadyLocked = true;			
								break;
							case sphere_lock:
								if (!ableToLockSphere(*iter, newLock))
									alreadyLocked = true;			
								break;
							default:
								throw ERROR_REPORT("Undefined custom lock type!");
						}
						break;
					// ---------------------------------------------------------
					default:
						throw ERROR_REPORT("Undefined custom lock type!");	
				}
			}
			
			if (alreadyLocked)
				break;
			
			++iter;			
		}
		
		if (!alreadyLocked)
		{
			// *** Lock acquired ***
			listOfCustomLocks.push_back(newLock);
			lockAcquired = true;
			
#ifdef MITOGEN_DEBUG_MULTITHREADING
			// get current time
			const report_clock::time_point currentTime = report_clock::now();	
			
			// get elapsed time
			const report_clock::duration timeToAcquireLock = currentTime - startTime;
			const std::chrono::milliseconds timeToAcquireLock_ms =
				std::chrono::duration_cast<std::chrono::milliseconds>(timeToAcquireLock);
			
			newLock->printLockInfo(timeToAcquireLock_ms);
			
			static report_clock::duration totalWaitingTime(0);
			totalWaitingTime += timeToAcquireLock;
			std::chrono::seconds totalWaitingTime_sec =
				std::chrono::duration_cast<std::chrono::seconds>(totalWaitingTime);
			std::chrono::seconds sec =
				std::chrono::duration_cast<std::chrono::seconds>
				(totalWaitingTime_sec % std::chrono::minutes(1));
			std::chrono::minutes min =
				std::chrono::duration_cast<std::chrono::minutes>(totalWaitingTime_sec);
				
			REPORT("Total time spent waiting (all threads): " << min.count() << "m " 
				<< sec.count() << "s" );
#endif
#ifdef GTGEN_WITH_LOCKVISUALIZATION
			// save visualization
			saveVisualization(newLock, true);
#endif
		}
		else
		{
			// *** Already locked, waiting for unlock ***
			justUnlocked.wait(listSearchLock);
			
			// *** Trying again ***
			// (search the list of locks again; while cycle continues)
		}			
	}
	
	iter = listOfCustomLocks.end();

	// end() returns the element past the end of the sequence
	return --iter;
}


bool SceneLocks::ableToLockTimePoints(CustomLock* curLock, CustomLock* newLock)
{
	for (size_t i = curLock->getFromTimePoint();
		i <= curLock->getToTimePoint(); ++i)
	{
		for (size_t j = newLock->getFromTimePoint();
			j <= newLock->getToTimePoint(); ++j)
		{
			// non-empty intersection of time points
			if (i == j)
				return false;
		}
	}
	return true;
}


bool SceneLocks::ableToLockSphere(CustomLock* curLock, CustomLock* newLock)
{
	SphereLock* _curLock = dynamic_cast<SphereLock*>(curLock);
	SphereLock* _newLock = dynamic_cast<SphereLock*>(newLock);

	for (size_t i = _curLock->getFromTimePoint();
		i <= _curLock->getToTimePoint(); ++i)
	{
		for (size_t j = _newLock->getFromTimePoint();
			j <= _newLock->getToTimePoint(); ++j)
		{
			// non-empty intersection of time points
			if (i == j)
			{
				// compute distance of sphere centres
				float distanceOfCentres = sqrt(
					  pow((_curLock->getSphereCentre().x - _newLock->getSphereCentre().x), 2)
					+ pow((_curLock->getSphereCentre().y - _newLock->getSphereCentre().y), 2)
					+ pow((_curLock->getSphereCentre().z - _newLock->getSphereCentre().z), 2));
				
				// minimal non-intersecting distance of spheres
				float minLockableDistance = _curLock->getSphereRadius()
					+ _newLock->getSphereRadius();
					
				if (distanceOfCentres < minLockableDistance)
					return false;						
			}
		}
	}
	return true;
}

#ifdef GTGEN_WITH_LOCKVISUALIZATION
void SceneLocks::saveVisualization(CustomLock* const newLock, const bool addedLock) const
{
	using namespace std;
	
	static size_t imgCounter = 0;
	
	// TODO: currently implemented only for sphere-locks on sceneMasks
	if (newLock->getLockType() != sphere_lock
		|| newLock->getSceneType() != masks)
		return;
	
	// draw only timePoints influenced by the current lock
	const size_t drawFromTimePoint = newLock->getFromTimePoint();
	const size_t drawToTimePoint = newLock->getToTimePoint();
	
	// prepare the image
	i3d::Image3d<i3d::GRAY8>* lockImg = new i3d::Image3d<i3d::GRAY8>();
	lockImg->SetResolution(this->imgRes);
	lockImg->MakeRoom(this->imgSize);
		
	// draw all current locks with corresponding timePoints
	for (size_t tP = drawFromTimePoint; tP <= drawToTimePoint; ++tP)
	{
		// how many sphere-locks are currently active
		int sphereCount = 0;
		
		// clear image
		lockImg->Clear();
		/*i3d::GRAY8 *c=lockImg->GetFirstVoxelAddr();
		for (size_t i=0; i < lockImg->GetImageSize(); ++i, ++c)
			*c=255;*/
		
		list<CustomLock*>::const_iterator iter = listOfCustomLocks.begin();		
		while (iter != listOfCustomLocks.end())
		{		
			if ((*iter)->getLockType() == sphere_lock
				&& (*iter)->getSceneType() == masks
				&& (*iter)->getFromTimePoint() <= tP
				&& (*iter)->getToTimePoint() >= tP) 
			{
				++sphereCount;
				
				SphereLock* _iterLock = dynamic_cast<SphereLock*>(*iter);

				// resolution
				const float xRes=lockImg->GetResolution().GetX();
				const float yRes=lockImg->GetResolution().GetY();
				const float zRes=lockImg->GetResolution().GetZ();
		
				// sphere
				const int x0=(int)floorf(_iterLock->getSphereCentre().x*xRes);
				const int y0=(int)floorf(_iterLock->getSphereCentre().y*yRes);
				const int z0=(int)floorf(_iterLock->getSphereCentre().z*zRes);
				const int r=(int)floorf(_iterLock->getSphereRadius()*xRes);
				
				//REPORT("Sphere centre: " << x0 << ", " << y0 << ", " << z0);

				// go through the image
				i3d::GRAY8 *m=lockImg->GetFirstVoxelAddr();
				const size_t cellID = _iterLock->getCellID();
				//REPORT("cell ID: " << cellID);
				for (size_t i=0; i < lockImg->GetImageSize(); ++i, ++m)
				{
					// get voxel coordinates
					int x = lockImg->GetX(i);
					int y = lockImg->GetY(i);
					int z = lockImg->GetZ(i);
		
					// write cellID if voxel is in the sphere
					// (cell IDs start from 1)
					if (pow((x-x0), 2) + pow((y-y0), 2) + pow((z-z0), 2) <= pow(r, 2))
						*m=cellID;
				}
			}

			++iter;
		}
		
		// get time stamp
		const report_clock::time_point currentTime_st = report_clock::now();
		const report_clock::duration runTime_st = currentTime_st - this->startTime;
	
		const chrono::milliseconds runTime_msec_st =
			chrono::duration_cast<chrono::milliseconds>(runTime_st); 
		const chrono::hours hh =
			chrono::duration_cast<chrono::hours>(runTime_msec_st);
		const chrono::minutes mm =
			chrono::duration_cast<chrono::minutes>(runTime_msec_st % chrono::hours(1));
		const chrono::seconds ss =
			chrono::duration_cast<chrono::seconds> (runTime_msec_st % chrono::minutes(1));
		const chrono::milliseconds msec =
			chrono::duration_cast<chrono::milliseconds> (runTime_msec_st % chrono::seconds(1));
			
		// get date & time
		time_t rawtime;
		struct tm* timeinfo;
		time (&rawtime);
		timeinfo = localtime (&rawtime);
		const int day = timeinfo->tm_mday;
		const int month = timeinfo->tm_mon + 1;
		const int year = timeinfo->tm_year + 1900;
		const int hour = timeinfo->tm_hour;
		const int min = timeinfo->tm_min;
		const int sec = timeinfo->tm_sec;
		
		// was lock added or removed?
		string addedOrRemoved;
		if (addedLock)
			addedOrRemoved = "afterLockAdded";
		else
			addedOrRemoved = "afterLockRemoved";				
			
		// generate file name
		SphereLock* _newLock = dynamic_cast<SphereLock*>(newLock);
		stringstream fileName;
		fileName << "lockImg_masks_tp" << setfill('0')
			<< setw(4) << tP << "_"
			<< setw(8) << imgCounter++ << "_"
			<< setw(2) << hh.count() << "h-" 
			<< setw(2) << mm.count() << "m-"
			<< setw(2) << ss.count() << "s-"
   			<< setw(3) << msec.count() << "ms_"
   			<< "time" << hour << "h" << min << "m" << sec << "s_" 
   			<< "date" << year << "-" << month << "-" << day << "_"
   			<< sphereCount << "locks_"
   			<< "cellId" << _newLock->getCellID() << "_"
   			<< addedOrRemoved << "." << SUFFIX;
   		
   		// resample image to custom resolution
   		i3d::Image3d<i3d::GRAY8>* resampledLockImg = new i3d::Image3d<i3d::GRAY8>();
   		i3d::Resample(*lockImg, *resampledLockImg,
   			this->imgSize.x/4,
   			this->imgSize.y/4,
   			this->imgSize.z/4);

   		// save image to HDD
   		resampledLockImg->SaveImage(fileName.str().c_str());
   		
   		// free memory
   		delete resampledLockImg;   		
	}

	// free memory
	delete lockImg;
}
#endif

#endif
